/*
** Module   : DatetimeFormat.java
** Abstract : datetime formatting and editing support
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20130327 Initial version.
*/

package com.goldencode.p2j.ui.client.format;

import static com.goldencode.p2j.util.datetime.*;

import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * Display format for <code>datetime</code> variables.
 */
public class DatetimeFormat
extends DisplayFormat
{
   // TODO: already existing instances won't pick up the latest change
   //       when SESSION:DATE-FORMAT is called, but in the 4GL they will
   /** Order of date components for rendering. */
   protected byte[] dateCompOrder = datetime.getDateComponentOrder(); //TODO: implement this static method in datetime 

   /** 
    * Date components separator.
    * In addition to date, there are some hardcoded time separator:
    * <ul>
    *    <li>one <b>space</b> that separates the date and time parts;
    *    <li><b>colon</b> that separates hours from minutes and minutes from seconds;
    *    <li><b>dot</b> that separates the seconds and milliseconds.
    * </li>
    */
   protected char fmtSeparator;

   /**
    * Determines whether the separators will be drawn for the unknown value
    * (applies to editing mode only). 
    */
   protected boolean drawSeparators = true;

   /** <code>true</code> for a 4-digit year. */
   protected boolean fmtFullYear;

   /**
    * How date components are positioned in rendered string. This array is
    * directly indexed by the component's code constants (DAY etc.).
    */
   protected DatetimeComponent[] fmtDatetimeComps;

   /** Default width which should be reserved for a screen value. */
   private int fmtScreenWidth;

   /**
    * Constructs a datetime format from the given Progress format string.
    * 
    * @param   pformat
    *          Progress format string.
    * @throws  DisplayFormatParsingException
    *          When passed format is not a valid datetime format.
    */
   public DatetimeFormat(String pformat) 
   throws DisplayFormatParsingException
   {
      super(pformat);

      // for a quick implementation the following tests were taken from date(String pformat)
      // but the exact match was replaced with startsWith and the space separator between date 
      // and time was added to matching strings. The comments remains valid.
      if (pformat.startsWith("99/99/99 "))
      {
         fmtSeparator = '/';
         fmtFullYear = false;
      }
      else if (pformat.startsWith("99/99/9999 "))
      {
         fmtSeparator = '/';
         fmtFullYear = (dateCompOrder[2] == YEAR);
      }
      else if (pformat.startsWith("99/9999/99 "))
      {
         fmtSeparator = '/';
         fmtFullYear = (dateCompOrder[1] == YEAR);
      }
      else if (pformat.startsWith("9999/99/99 "))
      {
         fmtSeparator = '/';
         fmtFullYear = (dateCompOrder[0] == YEAR);
      }
      else if (pformat.startsWith("99-99-99 "))
      {
         fmtSeparator = '-';
         fmtFullYear = false;
      }
      else if (pformat.startsWith("99-99-9999 "))
      {
         fmtSeparator = '-';
         fmtFullYear = (dateCompOrder[2] == YEAR);
      }
      else if (pformat.startsWith("99-9999-99"))
      {
         fmtSeparator = '-';
         fmtFullYear = (dateCompOrder[1] == YEAR);
      }
      else if (pformat.startsWith("9999-99-99 "))
      {
         fmtSeparator = '-';
         fmtFullYear = (dateCompOrder[0] == YEAR);
      }
      else if (pformat.startsWith("99.99.99 "))
      {
         fmtSeparator = '.';
         fmtFullYear = false;
      }
      else if (pformat.startsWith("99.99.9999 "))
      {
         fmtSeparator = '.';
         fmtFullYear = (dateCompOrder[2] == YEAR);
      }
      else if (pformat.startsWith("99.9999.99 "))
      {
         fmtSeparator = '.';
         fmtFullYear = (dateCompOrder[1] == YEAR);
      }
      else if (pformat.startsWith("9999.99.99 "))
      {
         fmtSeparator = '.';
         fmtFullYear = (dateCompOrder[0] == YEAR);
      }
      else if (pformat.startsWith("999999 "))
      {
         fmtSeparator = '\0';
         fmtFullYear = false;
      }
      else if (pformat.startsWith("99999999 "))
      {
         fmtSeparator = '\0';
         fmtFullYear = true;
      }
      else
      {
         throw new DisplayFormatParsingException("Invalid date format: " + pformat);
      }

      fmtDatetimeComps = new DatetimeComponent[7];
      int offs = 0;
      for (int i = 0; i < fmtDatetimeComps.length; i++)
      {
         if (i < dateCompOrder.length)
         {
            // date part
            int compIndex = dateCompOrder[i];
            int length = (compIndex == YEAR && fmtFullYear) ? 4 : 2;
            fmtDatetimeComps[compIndex] = new DatetimeComponent(offs, length);
            offs += length;
         }
         else 
         {
            // time part
            int length = (i == MILLIS) ? 3 : 2;
            fmtDatetimeComps[i] = new DatetimeComponent(offs, length);
            offs += length;
         }
         
         if (fmtSeparator != '\0') // for date part
         {
            offs++;
         }
         if (i >= HOURS) // for time part
         {
            offs++;
         }
      }

      fmtScreenWidth = 6 /* date */ + 1 /* space */ + 12 /* time */;
      if (fmtSeparator != '\0')
      {
         fmtScreenWidth += 2;
      }
      if (fmtFullYear)
      {
         fmtScreenWidth += 2;
      }
   }

   /**
    * Returns screen width of the format.  The screen width is the number of
    * characters that will be rendered.  This is used as the default size for
    * an editor.
    * 
    * @return   The screen width of the format. 
    */
   @Override
   public int getScreenWidth()
   {
      return fmtScreenWidth;
   }

   /**
    * Returns a new presentation instance with a default initialized insert mode and 
    * cursor position.
    * 
    * @param   pres
    *          A Presentation instance which was constructed by this format. May be 
    *          <code>null</code>. This presentation instance is not changed in this method.
    * 
    * @return  New presentation instance.
    */
   @Override
   public Presentation clear(Presentation pres)
   {
      drawSeparators = false;

      DatetimeBuf buf = new DatetimeBuf();
      
      buf.compBufs = null;

      if (pres != null)
      {
         buf.presInsertMode = pres.presInsertMode;
      }
      
      return buf;
   }

   /**
    * Initializes a presentation for variable's value.
    * 
    * @param   var
    *          Variable which should be compatible with format.
    * 
    * @return  Data presentation
    * 
    * @throws  DisplayFormatCheckException
    *          Variable's value is incompatible with format.
    */
   @Override
   public Presentation fromVar(BaseDataType var)
   throws DisplayFormatCheckException
   {
      return new DatetimeBuf((datetime)var);
   }

   /**
    * Forces separators for the unknown datetime value to be drawn or not in the
    * subsequent value rendering (applies to editing mode only).
    *
    * @param drawSeparators
    *        If <code>true</code> then separators will be displayed for the
    *        unknown datetime value.
    */
   public void setDrawSeparators(boolean drawSeparators)
   {
      this.drawSeparators = drawSeparators;
   }

   /**
    * The simple structure which records what places are occupied in a screen
    * value by the datetime component.
    */
   private static class DatetimeComponent
   {
      /** Offset of this component in rendered string. */
      private int compOffset;

      /** Number of digits to render. */
      private int compLength;

      /**
       * Constructor.
       * 
       * @param   offset
       *          Offset of this component in rendered string.
       * @param   length
       *          Number of digits to render.
       */
      public DatetimeComponent(int offset, int length)
      {
         compOffset = offset;
         compLength = length;
      }

      /**
       * Returns offset of this component in rendered string.
       * 
       * @return  Offset of this component in rendered string.
       */
      public int getOffset()
      {
         return compOffset;
      }

      /**
       * Returns number of digits to render.
       * 
       * @return  Number of digits to render.
       */
      public int getLength()
      {
         return compLength;
      }
   }

   /**
    * Internal form of datetime data.
    */
   private class DatetimeBuf
   extends Presentation
   {
      /** Special value for an empty digit place. */
      private static final char SPACE = ' ';

      /**
       * Buffers for datetime component digits. May be <code>null</code> which means UNKNOWN 
       * value. Array element must contain a digit or <code>SPACE</code> which will be rendered 
       * as space.
       */
      private char[][] compBufs = null;

      /**
       * Constructs default presentation used for clear operation.
       */
      public DatetimeBuf()
      {
         presCursorPos = 0;
         presInsertMode = false;
         initClear();
      }

      /**
       * Constructs presentation initialized with given <code>datetime</code>.
       * 
       * @param   varDatetime
       *          Initial value.
       * 
       * @throws  DisplayFormatCheckException
       *          This value can't be displayed with the current format.
       */
      public DatetimeBuf(datetime varDatetime)
      throws DisplayFormatCheckException
      {
         if (varDatetime.isUnknown())
         {
            return;
         }

         int year   = varDatetime.getYear();
         int month  = varDatetime.getMonth();
         int day    = varDatetime.getDay();
         int hours  = varDatetime.getHours();
         int mins   = varDatetime.getMinutes();
         int secs   = varDatetime.getSeconds();
         int millis = varDatetime.getMilliseconds();
         
         if (!fmtFullYear)
         {
            if (!varDatetime.inY2KWindow())
            {
               throw new DisplayFormatCheckException(
                  "Value " +
                  varDatetime.toStringMessage() +
                  " cannot be displayed using " +
                  getFormatDef() +
                  ". Year must be in current century.");
            }

            year %= 100;
         }

         compBufs = initBuffers();
         assignComponent(compBufs[YEAR],    year);
         assignComponent(compBufs[MONTH],   month);
         assignComponent(compBufs[DAY],     day);
         assignComponent(compBufs[HOURS],   hours);
         assignComponent(compBufs[MINUTES], mins);
         assignComponent(compBufs[SECONDS], secs);
         assignComponent(compBufs[MILLIS],  millis);
      }

      /**
       * Checks format rules. Method finishEdit() should be called before it.
       * 
       * @throws  DisplayFormatCheckException
       *          When data is incompatible with format
       */
      @Override
      public void checkFormat()
      throws DisplayFormatCheckException
      {
         // just call this to check for exceptions, discard the return value
         toDatetime();
      }

      /**
       * Validate representation according to format rules and return error
       * message if there is an error.
       * 
       * @return   <code>null</code> if no error or error message if there is an
       *           error.
       */
      @Override
      public ErrorEntry validateFormat()
      {
         int day     = -1;
         int month   = -1;
         int hours   = -1;
         int minutes = -1;
         int seconds = -1;
         int millis  = -1;
         
         // assume unedited datetime is valid
         if (compBufs == null)
            return null;
         
         try
         {
            day     = extractDayOrMonthComponent(compBufs[DAY]);
            month   = extractDayOrMonthComponent(compBufs[MONTH]);
            hours   = extractTimeComponent(compBufs[HOURS]);
            minutes = extractTimeComponent(compBufs[MINUTES]);
            seconds = extractTimeComponent(compBufs[SECONDS]);
            millis  = extractTimeComponent(compBufs[MILLIS]);
         }
         catch (NumberFormatException e)
         {
            // error will be handled later
         }

         ErrorEntry res = verifyYearComponent(compBufs);
         if (res != null)
         {
            return res;
         }

         try
         {
            int year = extractYearComponent(compBufs);
            return datetime.validate(month, day, year, !fmtFullYear,
                                     hours, minutes, seconds, millis);
         }
         catch (ErrorConditionException ex)
         {
            // should not happen because we already verified data
         }
         
         return null;
      }
      
      /**
       * Returns if the presentation currently represents the 
       * <code>unknown value</code>.  If any editing has occurred, then this
       * will return <code>false</code> even if the editing is not complete.
       *
       * @return   <code>true</code> if this is <code>unknown</code>.
       */
      @Override
      public boolean isUnknown()
      {
         return (compBufs == null);
      }

      /**
       * Erases one symbol before the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteBack()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
            return true;
         }
         
         return false;
      }

      /**
       * Erases one symbol in the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteCurrent()
      {
         return false;
      }

      /**
       * Inputs character. The result depends from the current insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return  <code>true</code> if input is accepted (nevertheless it may be silently 
       *          ignored in reality).
       */
      @Override
      public boolean input(char ch)
      {
         if (presCursorPos >= getScreenWidth())
         { 
            // ignore any input outside widget even '?'
            return false;
         }

         if (ch == '?')
         {
            compBufs = null;
            presCursorPos = 0;
            drawSeparators = false;
            return true;
         }

         if (compBufs == null)
         { 
            // restart method from zero content
            initClear();
            boolean res = input(ch);
            if (!res)
            {
               compBufs = null;
            }
            return res;
         }

         int cursorPos = presCursorPos;
         
         // clear flag for last char
         setLastCharEntered(false);
         
         // iterate through datetime components
         for (int i = 0; i < dateCompOrder.length; i++)
         {
            byte compIndex = dateCompOrder[i];
            DatetimeComponent comp = fmtDatetimeComps[compIndex];
            int offset = comp.getOffset();

            // bypass components which we are not in
            if (cursorPos >= offset + comp.getLength())
            {
               continue;
            }

            byte digit = (byte) Character.digit(ch, 10);
            int pos = cursorPos - offset;
            if (digit >= 0)
            {
               if (pos < 0)
               {
                  // skip separator
                  pos = 0;
                  cursorPos = offset;
               }

               compBufs[compIndex][pos] = ch;

               // is this the last digit position of the last component?
               if (i == dateCompOrder.length - 1 && pos == compBufs[compIndex].length - 1)
               {
                  setLastCharEntered(true);
               }

               break;
            }

            if (ch == '-' && compIndex == YEAR  && pos >= 0 && fmtSeparator != 0)
            {
               compBufs[compIndex][pos] = ch;

               // is this the last digit position of the last component?
               if (i == dateCompOrder.length - 1 && pos == compBufs[compIndex].length - 1)
               {
                  setLastCharEntered(true);
               }

               break;
            }

            if (ch == fmtSeparator)
            {
               if (pos < 0)
               {
                  // Separator char entered over existing separator.
                  // Only moving cursor
                  break;
               }

               // completing current component and moving to next if possible
               completeComponent(compBufs, compBufs, compIndex);
               if (i < dateCompOrder.length - 1)
               {
                  // next component available. Separator is always present in
                  // this case so we can calculate offset with only current
                  // component
                  presCursorPos = offset + comp.getLength() + 1;
               }
               else
               { 
                  // no move
                  presCursorPos = cursorPos;
               }
               return true;
            }

            // invalid input rings the terminal bell
            OutputManager.instance().beep();
            
            return false;
         }

         // move the cursor right even if we go past the end by 1
         presCursorPos = cursorPos + 1;

         return true;
      }

      /**
       * Moves cursor one character to the left.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorLeft()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
         }
         return true;
      }

      /**
       * Moves cursor one character to the right. Also updates insert mode.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorRight()
      {
         if (presCursorPos < getScreenWidth() - 1)
         {
            presCursorPos++;
         }
         return true;
      }

      /**
       * Moves cursor to the beginning of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorBegin()
      {
         presCursorPos = 0;
         return true;
      }

      /**
       * Moves cursor to the end of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorEnd()
      {
         presCursorPos = getScreenWidth() - 1;
         return true;
      }

      /**
       * Indicates completion of an edit. The presentation instance may fill 
       * some unfilled places.
       * 
       * @return   <code>true</code> if rendering is necessary.
       */
      @Override
      public boolean finishEdit()
      {
         if (compBufs == null)
         {
            return false;
         }

         char[][] newBufs = initBuffers();
         boolean changed = false;
         
         for (int i = 0; i < newBufs.length; i++)
         {
            completeComponent(newBufs, compBufs, i);
            if (!Arrays.equals(newBufs[i], compBufs[i]))
            {
               changed = true;
            }
         }

         if (changed)
         {
            compBufs = newBufs;
         }

         return changed;
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @param    edit
       *           <code>true</code> if the screen value is being generated
       *           for interactive editing purposes.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue(boolean edit)
      {
         boolean tempBuf = (edit && (compBufs == null) && drawSeparators);

         if (tempBuf)
         {
            // create temporary buffers
            initClear();
         }
         
         String result = toScreenValue();
         
         if (tempBuf)
         {
            // remove temporary buffers
            compBufs = null;
         }
         
         return result;
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue()
      {
         if (compBufs == null)
         {
            return "";
         }

         StringBuffer res = new StringBuffer(10);
         for (int i = 0; i < dateCompOrder.length; i++)
         {
            if (i > 0 && i < HOURS && fmtSeparator != '\0')
            {
               res.append(fmtSeparator);
            }
            if (i == HOURS)
            {
               res.append(' '); //separates date and time
            }

            res.append(compBufs[dateCompOrder[i]]);
            
            switch (i)
            {
               case MILLIS:
                  break;
               case SECONDS:
                  res.append('.');
                  break;
               case HOURS:
               case MINUTES:
                  res.append(':');
                  break;
            }
         }

         return res.toString();
      }

      /**
       * Instantiates a variable that is equivalent to the presentation's
       * internal representation.
       * 
       * @param    var
       *           A variable which must be compatible with the format.
       */
      @Override
      public void toVar(BaseDataType var)
      {
         try
         {
            ((datetime) var).assign(toDatetime(), true);
         }
         catch (DisplayFormatCheckException ex)
         {
            // buffers should be already validated
         }
      }

      /**
       * Allocates datetime components buffers.
       * 
       * @return  Components buffers array.
       */
      private char[][] initBuffers()
      {
         char[][] bufs = new char[7][];
         bufs[YEAR]    = new char[(fmtFullYear) ? 4 : 2];
         bufs[MONTH]   = new char[2];
         bufs[DAY]     = new char[2];
         bufs[HOURS]   = new char[2];
         bufs[MINUTES] = new char[2];
         bufs[SECONDS] = new char[2];
         bufs[MILLIS]  = new char[3];
         return bufs;
      }

      /**
       * Resets content to space value.
       */
      private void initClear()
      {
         compBufs = initBuffers();
         for (int i = 0; i < compBufs.length; i++)
         {
            char[] compBuf = compBufs[i];
            clearComponent(compBuf);
         }
      }

      /**
       * Fills the date component with <code>SPACE</code>.
       * 
       * @param   comp
       *          The component.
       */
      private void clearComponent(char[] comp)
      {
         for (int i = 0; i < comp.length; ++i)
         {
            comp[i] = SPACE;
         }
      }

      /**
       * Fills the date component with the given value.
       * 
       * @param   comp
       *          The component.
       * @param   value
       *          The new value.
       */
      private void assignComponent(char[] comp, int value)
      {
         String txt = Integer.toString(value);
         
         int idx = txt.length() - 1;
         
         for (int i = comp.length - 1; i >= 0; i--)
         {
            if (idx >= 0)
            {
               comp[i] = txt.charAt(idx);
               idx--;
            }
            else
            {
               // left pad
               comp[i] = '0';
            }
         }
      }

      /**
       * Calculates numeric value of the date component.
       * 
       * @param   comp
       *          The date component (day or month). May partially consist from spaces.
       *
       * @return  Not negative integer value of component. (-1) if it is empty.
       */
      private int extractDayOrMonthComponent(char[] comp)
      {
         int begin = 0;

         while (begin < comp.length)
         {
            if (comp[begin] != SPACE)
            {
               break;
            }
            begin++;
         }

         if (begin == comp.length)
         {
            return -1;
         }

         int end = begin + 1;
         while (end < comp.length)
         {
            if (comp[end] == SPACE)
            {
               break;
            }
            end++;
         }

         String txt = new String(comp);
         
         return Integer.parseInt(txt.substring(begin, end));
      }

      /**
       * Calculates numeric value of the time component.
       * 
       * @param   comp
       *          The time component (hours/mins/secs/millis).
       *
       * @return  Not negative integer value of component. (-1) if it is empty.
       */
      private int extractTimeComponent(char[] comp)
      {
         return Integer.parseInt(new String(comp));
      }

      /**
       * Calculates numeric value of the year. Calculation can be described by
       * the following algorithm:
       * <p>
       * If format separator isn't <code>null</code> then:
       * <p>
       * <ol>
       *    <li> Input sequence is autocompleted (to a temporary buffer).
       *    <li> If the extreme right position holds '-' then it is discarded
       *         (e.g. -10- -> -10)
       *    <li> Remaining characters are parsed as a usual integer.
       * </ol>
       * <p>
       * If format separator is <code>null</code> then:
       * <p>
       * <ol>
       *    <li> Only the extreme left (before the space if it presents)
       *         digits are taken into account. If the first character is a
       *         space then the whole field content is ignored (e.g. 12_3 ->
       *         12; _123 -> <code>null</code>)
       *    <li> If input consists of three digits then the last one is
       *         ignored (e.g. 123_ -> 12).
       *    <li> Remaining characters are parsed as a usual integer. If it
       *         cannot be parsed then the current year is used.
       * </ol>
       *
       * @param   comp
       *          The input datetime buffer.
       *
       * @return  The numeric value of the year.
       *
       * @throws  ErrorConditionException
       *          If the input sequence does not represent a valid year.
       */
      private int extractYearComponent(char[][] comp)
      throws ErrorConditionException
      {
         char[][] tempBuf = new char[3][];
         tempBuf[YEAR] = Arrays.copyOf(comp[YEAR], comp[YEAR].length);
         for (char c : comp[YEAR])
         {
            if (c == SPACE)
            {
               // process autocompletion for a temporary buffer if it is required
               completeComponent(tempBuf, comp, YEAR);
               break;
            }
         }

         char[] yearBuffer = tempBuf[YEAR];

         int year = 0;
         if (fmtSeparator != 0)
         {
            int end = yearBuffer.length - 1;

            // discard extreme right '-' if present
            if (yearBuffer[end] == '-')
            {
               end--;
            }

            String txt = new String(yearBuffer);
            try
            {
               year = Integer.parseInt(txt.substring(0, end + 1));
            }
            catch (NumberFormatException ex)
            {
               if (fmtFullYear)
               {
                  ErrorManager.recordOrThrowError(85, "Invalid datetime input");
                  return (int)datetime.INVALID_DATE;
               }
            }
         }
         else // fmtSeparator == 0
         {
            // Only extreme left (before the space if it presents) digits
            // are taken into account. If the first character is a space
            // then the whole field content is ignored.
            int end = yearBuffer.length - 1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] == SPACE)
               {
                  end = i - 1;
                  break;
               }
            }

            String txt = new String(yearBuffer);
            txt = txt.substring(0, end + 1);

            // if input consists of three digits then the last one is ignored
            if (txt.length() == 3)
            {
               txt = txt.substring(0, 2);
            }

            try
            {
               year = Integer.parseInt(txt);
            }
            catch (NumberFormatException ex)
            {
               year = (new date()).getYear();
            }
         }

         if (year == 0 && fmtFullYear)
         {
            ErrorManager.recordOrThrowError(79, "Year is out of range of 0");
            return (int) datetime.INVALID_DATE;
         }

         return year;
      }

      /**
       * Validate year without throwing exceptions.
       * 
       * @param    comp
       *           Component to verify.
       * 
       * @return   <code>null</code> if no error found, {@link ErrorEntry}
       *           instance otherwise.
       */
      private ErrorEntry verifyYearComponent(char[][] comp)
      {
         char[][] tempBuf = new char[3][];
         tempBuf[YEAR] = Arrays.copyOf(comp[YEAR], comp[YEAR].length);
         for (char c : comp[YEAR])
         {
            if (c == SPACE)
            {
               // process autocompletion for a temporary buffer if it is required
               completeComponent(tempBuf, comp, YEAR);
               break;
            }
         }

         char[] yearBuffer = tempBuf[YEAR];

         int year = 0;
         if (fmtSeparator != 0)
         {
            int end = yearBuffer.length - 1;

            // discard extreme right '-' if present
            if (yearBuffer[end] == '-')
            {
               end--;
            }

            String txt = new String(yearBuffer);
            try
            {
               year = Integer.parseInt(txt.substring(0, end + 1));
            }
            catch (NumberFormatException ex)
            {
               if (fmtFullYear)
               {
                  return new ErrorEntry(85, "Invalid datetime input");
               }
            }
         }
         else // fmtSeparator == 0
         {
            // Only extreme left (before the space if it presents) digits
            // are taken into account. If the first character is a space
            // then the whole field content is ignored.
            int end = yearBuffer.length - 1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] == SPACE)
               {
                  end = i - 1;
                  break;
               }
            }

            String txt = new String(yearBuffer);
            txt = txt.substring(0, end + 1);

            // if input consists of three digits then the last one is ignored
            if (txt.length() == 3)
            {
               txt = txt.substring(0, 2);
            }

            try
            {
               year = Integer.parseInt(txt);
            }
            catch(NumberFormatException ex)
            {
               year = (new date()).getYear();
            }
         }

         if (year == 0 && fmtFullYear)
         {
            return new ErrorEntry(79, "Year is out of range of 0");
         }

         return null;
      }
      
      /**
       * Completes datetime component. Day and month are completed with 0's.
       * <p>
       * This method uses only one old and one new buffer for component but
       * takes all buffers to simplify coding of its call.
       * <p>
       * The year completion algorithm can be described as follows:
       * <p>
       * <ol>
       *    <li> If format separator is <code>null</code> then do nothing.
       *    <li> Find the extreme right '-' and fill all empty placeholders
       *         to the left of it with the '-' (e.g. __-_ -> ---_; _1_- ->
       *         -1--).
       *    <li> The extreme leftmost meaningful symbol (not SPACE) is found.
       *         After that we search for the extreme right meaningful symbol
       *         such that no spaces are present between it and the extreme
       *         left meaningful symbol. All symbols which are not between
       *         these ones are discarded (e.g. __12 -> 12; -2_3 -> -2;
       *         _1_3 -> 1). If no meaningful symbols are present then the
       *         current year is used.
       *    <li> To the previously found sequence of meaningful symbols some
       *         leading symbol may be added depending on the sequence length
       *         and year format:
       *    <ul>
       *       <li> When length is 3 and format is full-year OR
       *            length is 1 and format is not full-year then add a
       *            leading 0.
       *       <li> When length is 2 or 1 and format is full-year then the
       *            date.windowingImpl() which takes as input the digits from
       *            the sequence (the one that was created on the third step)
       *            is used (if the sequence contains no digits then 0 is
       *            used).
       *       <li> From the return value we take first 2 or 3 characters and
       *            complete the sequence with it. (e.g. -2- -> 0-2-; 99 ->
       *            1999; -- -> 20--; 9 -> 2009)
       *     </ul>
       * </ol>
       *
       * @param   newCompBufs
       *          Buffers for new value. May be the same with <code>oldCompBufs</code>.
       * @param   oldCompBufs
       *          Buffers for old value.
       * @param   compIndex
       *          Component's index (DAY, MONTH, YEAR, HOURS, MINUTES, SECONDS, MILLIS).
       */
      private void completeComponent(char[][] newCompBufs,
                                     char[][] oldCompBufs,
                                     int      compIndex)
      {
         if (compIndex != YEAR)
         {
            int comp = extractDayOrMonthComponent(oldCompBufs[compIndex]);
            if (comp < 0)
            {
               comp = 0;
            }

            assignComponent(newCompBufs[compIndex], comp);
         }
         else
         {
            // compIndex == YEAR
            
            if (fmtSeparator == 0)
            {
               newCompBufs[compIndex] = Arrays.copyOf(oldCompBufs[compIndex],
                                                      oldCompBufs[compIndex].length);
               return;
            }

            char[] yearBuffer = Arrays.copyOf(oldCompBufs[compIndex],
                                              oldCompBufs[compIndex].length);

            // find extreme right '-'
            int rightMinusPos = -1;
            for (int i = yearBuffer.length - 1; i >= 0; i--)
            {
               if (yearBuffer[i] == '-')
               {
                  rightMinusPos = i;
                  break;
               }
            }

            if (rightMinusPos != -1)
            {
               // fill all empty placeholders to the left of the
               // extreme right '-' with '-'

               for (int i = rightMinusPos - 1; i >= 0; i--)
               {
                  if (yearBuffer[i] == SPACE)
                  {
                     yearBuffer[i] = '-';
                  }
               }
            }

            // find extreme left meaningful symbol (not SPACE)
            int leftSymbolPos = -1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] != SPACE)
               {
                  leftSymbolPos = i;
                  break;
               }
            }

            if (leftSymbolPos == -1)
            {
               // empty year - fill with the current one
               assignComponent(newCompBufs[compIndex],
                               (new date()).getYear());
               return;
            }

            // find extreme right meaningful symbol between which and the
            // extreme left meaningful symbol no spaces present
            int rightSymbolPos = leftSymbolPos;
            for (int i = leftSymbolPos + 1; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] != SPACE)
               {
                  rightSymbolPos = i;
               }
               else
               {
                  break;
               }
            }

            // length of the meaningful symbols sequence
            // all other symbols which are not present in this sequence
            // will not be taken into account any further
            int symbolsLength = rightSymbolPos - leftSymbolPos + 1;

            // move meaningful symbols to the right
            int writePos = yearBuffer.length - 1;
            for (int readPos = rightSymbolPos;
                 readPos >= leftSymbolPos;
                 readPos--)
            {
               yearBuffer[writePos--] = yearBuffer[readPos];
            }

            if (fmtFullYear)
            {
               // add leading year numbers
               if (symbolsLength == 3)
               {
                  yearBuffer[0] = '0';
               }
               else if (symbolsLength == 2 || symbolsLength == 1)
               {
                  // add two leading year numbers
                  StringBuffer lastYearDigits = new StringBuffer("");
                  if (symbolsLength == 2 && Character.isDigit(yearBuffer[2]))
                  {
                     lastYearDigits.append(yearBuffer[2]);
                  }
                  if (Character.isDigit(yearBuffer[3]))
                  {
                     lastYearDigits.append(yearBuffer[3]);
                  }

                  int lastYearDigitsInt = 0;
                  if (lastYearDigits.length() > 0)
                  {
                     lastYearDigitsInt = Integer.parseInt(lastYearDigits.toString());
                  }

                  String fullYear = Integer.toString(date.windowingImpl(lastYearDigitsInt));

                  if (fullYear.length() == 4)
                  {
                     for (int i = 0; i < yearBuffer.length - symbolsLength; i++)
                     {
                        yearBuffer[i] = fullYear.charAt(i);
                     }
                  }
                  else
                  {
                     // just in case
                     for (int i = 0; i < yearBuffer.length - symbolsLength; i++)
                     {
                        yearBuffer[i] = '0';
                     }
                  }
               }
            }
            else // not fmtFullYear
            {
               if (symbolsLength == 1)
               {
                  yearBuffer[0] = '0';
               }
            }

            newCompBufs[compIndex] = Arrays.copyOf(yearBuffer, yearBuffer.length);
         }
      }

      /**
       * Returns the current presentation's value converted to <code>datetime</code>.
       * 
       * @return  The current presentation's value converted to <code>datetime</code>.
       * 
       * @throws  DisplayFormatCheckException
       *          Value doesn't represent a correct datetime.
       */
      private datetime toDatetime()
      throws DisplayFormatCheckException
      {
         if (compBufs == null)
         {
            return datetime.instantiateUnknownDatetime();
         }

         // just extracting components. Spaces should be already eliminated
         try
         {
            int year = extractYearComponent(compBufs);
            if (!fmtFullYear)
            {
               year = date.windowingImpl(year);
            }
            int month = extractDayOrMonthComponent(compBufs[MONTH]);
            int day   = extractDayOrMonthComponent(compBufs[DAY]);
            int hours = extractTimeComponent(compBufs[HOURS]);
            int mins  = extractTimeComponent(compBufs[MINUTES]);
            int secs  = extractTimeComponent(compBufs[SECONDS]);
            int millis= extractTimeComponent(compBufs[MILLIS]);
            
            return new datetime(month, day, year, hours, mins, secs, millis);
         }
         catch (ErrorConditionException ex)
         {
            throw new DisplayFormatCheckException(ex.getMessage());
         }
      }
   }
}
