/*
** Module   : datetimetz.java
** Abstract : Progress 4GL compatible datetime-tz object
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------------
** 001 GES 20130128 Initial commit with basic functionality.
** 002 GES 20130322 Fixed the BDT c'tor and modified the assign(BDT,boolean) to be the
**                  common location for the _POLY morphing. This is safe because all current
**                  usage would have generated a ClassCastException if that method was ever used
**                  with an invalid type. Now, instead of the CCE, the code will morph the value
**                  as is possible. The morphing code still needs to be finished.
** 003 OM  20130329 Added time-zone components constants for datetime-tz formatter. Added 
**                  components validation function.
*/

package com.goldencode.p2j.util;

import java.util.*;

import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * A class that represents a Progress 4GL compatible datetime-tz object whose
 * data is mutable.  All Progress language features related to date data
 * types are supported including all function that can be accessed via the
 * date aware operators and built-in Progress date-related functions.
 * <p>
 * This class extends the <code>datetime</code> class so all methods should work
 * directly. If they don't they will be overridden. Also should the following
 * static implementations:
 * <p>
 * <pre>
 * now           function (as initializer for datetime-tz)     {@link #datetimetz()}
 * datetime-tz   function (date, integer)                      {@link #datetimetz(date, NumberType)}
 * datetime-tz   function (datetime, integer)                  {@link #datetimetz(date, NumberType)}
 * datetime-tz   function (datetime-tz, integer)               {@link #datetimetz(date, NumberType)}
 * datetime-tz   function (date, integer, integer)             {@link #datetimetz(date, NumberType, NumberType)}
 * datetime-tz   function (integer m, d, y, h, mm, s, ms, tz)  {@link #datetimetz(NumberType, NumberType, NumberType, NumberType, NumberType, NumberType, NumberType, NumberType)}
 * datetime-tz   function (string parameter)                   {@link #datetimetz(String)}
 * timezone      function (datetime-tz parameter)              {@link #getTimeZoneOffset}
 * </pre>
 * <p>
 * Default data type display format: 99/99/9999 HH:MM:SS.SSS+HH:MM
 */
public class datetimetz
extends datetime
{
   /** Represents the hours time zone component of the string formatted datetime-tz. */
   public static final byte TZ_HOURS = 7;

   /** Represents the minutes time zone component of the string formatted datetime-tz. */
   public static final byte TZ_MINS = 8;
   
   /** The default format string. */
   private static final String defaultFormat = "99/99/9999 HH:MM:SS.sss+HH:MM";

   /** Time zone offset in minutes from Coordinated Universal Time (UTC). */
   private int offset = INVALID_TZ;
   
   /**
    * Default constructor which is the equivalent of the built-in 4GL function
    * <code>NOW</code>.
    */
   public datetimetz()
   {
      super();
      
      // the timezone should already be set to the default
      offset = getDefaultOffset();
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType,boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public datetimetz(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         assign(value, false);
      }
   }
   
   /**
    * Constructs an instance with the exact same state as the passed
    * instance.  This can be used both as a copy constructor as well as an
    * implementation of the DATETIME-TZ() built-in function when called with a
    * <code>date</code> subclass.
    *
    * @param    d
    *           The instance from which to copy state.
    */
   public datetimetz(date d)
   {
      super(d); // assign date (and time if the given date is a datetime)

      if (d instanceof datetimetz)
      {
         // the timezone should already be set, now save the offset, if needed
         this.offset = ((datetimetz)d).offset;
      }
   }

   /**
    * Constructs an instance using the inputs, with varying behavior depending
    * on the type of the first parameter.
    * <p>
    * If the first parameter is a <code>date</code> instance, then the date is
    * copied from that instance and the second parameter is treated as the
    * time value (number of milliseconds since midnight).
    * <p>
    * If the first parameter is a <code>datetime</code> or 
    * <code>datetimetz</code> instance, then the date and time is copied from
    * the instance and the second parameter is treated as a positive or
    * negative timezone offset for the new instance.
    * <p>
    * This is an implementation of 3 forms of the DATETIME-TZ() built-in
    * function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the default timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * <p>
    * Any input that is <code>unknown</code> will cause this instance to be
    * <code>unknown</code>.
    * 
    * @param    d
    *           The date (or date subclass) from which to copy.
    * @param    timeOrOffset
    *           The milliseconds since midnight if the first parameter is a
    *           <code>date</code> or a timezone offset from UTC in minutes
    *           if the first parameter is a <code>datetime</code>. As an
    *           offset, the value must be between -840 and 840.
    */
   public datetimetz(date d, NumberType timeOrOffset)
   {
      if (d.isUnknown() || timeOrOffset.isUnknown())
      {
         setUnknown();
      }
      else
      {
         if (d instanceof datetime)
         {
            datetime dt = (datetime)d;
            
            // this path is for the datetime and datetimetz
            setOffset(timeOrOffset.intValue());
            
            instantiateFromDateTime(new date(d),
                                    dt.getTime(),
                                    getZoneFromOffset(offset));
         }
         else
         {
            // this path is for the date superclass
            setOffset(getDefaultOffset());
            
            instantiateFromDateTime(d,
                                    timeOrOffset.intValue(),
                                    getDefaultTimeZone());
         }
      }
   }

   /**
    * Constructs an instance using the inputs, with varying behavior depending
    * on the type of the first parameter.
    * <p>
    * If the first parameter is a <code>date</code> instance, then the date is
    * copied from that instance and the second parameter is treated as the
    * time value (number of milliseconds since midnight).
    * <p>
    * If the first parameter is a <code>datetime</code> or 
    * <code>datetimetz</code> instance, then the date and time is copied from
    * the instance and the second parameter is treated as a positive or
    * negative timezone offset for the new instance.
    * <p>
    * This is an implementation of 3 forms of the DATETIME-TZ() built-in
    * function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the default timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * <p>
    * Any input that is <code>unknown</code> will cause this instance to be
    * <code>unknown</code>.
    * 
    * @param    d
    *           The date (or date subclass) from which to copy.
    * @param    timeOrOffset
    *           The milliseconds since midnight if the first parameter is a
    *           <code>date</code> or a timezone offset from UTC in minutes
    *           if the first parameter is a <code>datetime</code>.  As an
    *           offset, the value must be between -840 and 840.
    */
   public datetimetz(date d, double timeOrOffset)
   {
      this(d, new integer((int) timeOrOffset));
   }
   
   /**
    * Constructs an instance using the inputs.
    * <p>
    * This is an implementation of 1 form of the DATETIME-TZ() built-in
    * function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the given timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * <p>
    * Any input that is <code>unknown</code> will cause this instance to be
    * <code>unknown</code>.
    * 
    * @param    d
    *           The date from which to copy.  This must be an instance of the
    *           <code>date</code> superclass and may NOT be an instance of the
    *           <code>datetime</code> or <code>datetimetz</code> since that is
    *           a 4GL error.
    * @param    time
    *           The milliseconds since midnight.
    * @param    offset
    *           The timezone offset from UTC in minutes. Must be between -840
    *           and 840 inclusive.
    */
   public datetimetz(date d, NumberType time, NumberType offset)
   {
      if (d.isUnknown() || time.isUnknown() || offset.isUnknown())
      {
         setUnknown();
      }
      else
      {
         if (d instanceof datetime)
         {
            // TODO: record or throw an error here
         }
         else
         {
            int tzoffset = offset.intValue();
            
            // this path is for the date superclass
            setOffset(tzoffset);
            
            instantiateFromDateTime(d,
                                    time.intValue(),
                                    getZoneFromOffset(tzoffset));
         }
      }
   }
   
   /**
    * Constructs an instance using the inputs.
    * <p>
    * This is an implementation of 1 form of the DATETIME-TZ() built-in
    * function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the given timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * <p>
    * Any input that is <code>unknown</code> will cause this instance to be
    * <code>unknown</code>.
    * 
    * @param    d
    *           The date from which to copy.  This must be an instance of the
    *           <code>date</code> superclass and may NOT be an instance of the
    *           <code>datetime</code> or <code>datetimetz</code> since that is
    *           a 4GL error.
    * @param    time
    *           The milliseconds since midnight.
    * @param    offset
    *           The timezone offset from UTC in minutes. Must be between -840
    *           and 840 inclusive.
    */
   public datetimetz(date d, NumberType time, double offset)
   {
      this(d, time, new integer((int) offset));
   }
   
   /**
    * Constructs an instance using the inputs.
    * <p>
    * This is an implementation of 1 form of the DATETIME-TZ() built-in
    * function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the given timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * <p>
    * Any input that is <code>unknown</code> will cause this instance to be
    * <code>unknown</code>.
    * 
    * @param    d
    *           The date from which to copy.  This must be an instance of the
    *           <code>date</code> superclass and may NOT be an instance of the
    *           <code>datetime</code> or <code>datetimetz</code> since that is
    *           a 4GL error.
    * @param    time
    *           The milliseconds since midnight.
    * @param    offset
    *           The timezone offset from UTC in minutes. Must be between -840
    *           and 840 inclusive.
    */
   public datetimetz(date d, double time, NumberType offset)
   {
      this(d, new integer((int) time), offset);
   }
   
   /**
    * Constructs an instance using the inputs.
    * <p>
    * This is an implementation of 1 form of the DATETIME-TZ() built-in
    * function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the given timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * <p>
    * Any input that is <code>unknown</code> will cause this instance to be
    * <code>unknown</code>.
    * 
    * @param    d
    *           The date from which to copy.  This must be an instance of the
    *           <code>date</code> superclass and may NOT be an instance of the
    *           <code>datetime</code> or <code>datetimetz</code> since that is
    *           a 4GL error.
    * @param    time
    *           The milliseconds since midnight.
    * @param    offset
    *           The timezone offset from UTC in minutes. Must be between -840
    *           and 840 inclusive.
    */
   public datetimetz(date d, double time, double offset)
   {
      this(d, new integer((int) time), new integer((int) offset));
   }
   
   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    */
   public datetimetz(double month,
                     double day,
                     double year,
                     double hours,
                     double minutes,
                     double seconds,
                     double millis)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime((int) hours, (int) minutes, (int) seconds, (int) millis);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    *           The offset in minutes from UTC. Must be between -840 and 840
    *           inclusive.
    */
   public datetimetz(double month,
                     double day,
                     double year,
                     double hours,
                     double minutes,
                     double seconds,
                     double millis,
                     double offset)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime((int) hours,
              (int) minutes,
              (int) seconds,
              (int) millis,
              (int) offset);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    */
   public datetimetz(double month,
                     double day,
                     double year,
                     double hours,
                     double minutes,
                     double seconds)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime((int) hours, (int) minutes, (int) seconds, 0);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    */
   public datetimetz(double month,
                     double day,
                     double year,
                     double hours,
                     double minutes)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime((int) hours, (int) minutes, 0, 0);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function. Any
    * component that is <code>unknown</code> will result in the instance being
    * <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    */
   public datetimetz(NumberType month,
                     NumberType day,
                     NumberType year,
                     NumberType hours,
                     NumberType minutes,
                     NumberType seconds,
                     NumberType millis)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // use the version of setTime that handles unknown properly
      setTime(hours, minutes, seconds, millis);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function. Any
    * component that is <code>unknown</code> will result in the instance being
    * <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    * @param    offset
    *           The offset in minutes from UTC. Must be between -840 and 840
    *           inclusive.
    */
   public datetimetz(NumberType month,
                     NumberType day,
                     NumberType year,
                     NumberType hours,
                     NumberType minutes,
                     NumberType seconds,
                     NumberType millis,
                     NumberType offset)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // use the version of setTime that handles unknown properly
      setTime(hours, minutes, seconds, millis, offset);
   }

   /**
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function. Any
    * component that is <code>unknown</code> will result in the instance being
    * <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    */
   public datetimetz(NumberType month,
                     NumberType day,
                     NumberType year,
                     NumberType hours,
                     NumberType minutes,
                     NumberType seconds)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // use the version of setTime that handles unknown properly
      setTime(hours, minutes, seconds, new integer(0));
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME-TZ() built-in function. Any
    * component that is <code>unknown</code> will result in the instance being
    * <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    */
   public datetimetz(NumberType month,
                     NumberType day,
                     NumberType year,
                     NumberType hours,
                     NumberType minutes)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // use the version of setTime that handles unknown properly
      setTime(hours, minutes, new integer(0), new integer(0));
   }

   /**
    * Creates a new instance from the given date and time specification. This
    * is one form of the DATETIME-TZ built-in function. This is also the form of
    * the constructor that is used to handle initializer expressions (literals).
    * <p>
    * The specification must match the following format rules:
    * <p>
    * <ol>
    *   <li> Date value (this is not optional).  The string cannot start with
    *        whitespace. The first input is a date value that has at least a
    *        month and day specified or a month, day and year. The values
    *        are parsed in order as specified in SESSION:DATE-FORMAT. Zero
    *        padding is allowed but not required.
    *   <li> Date and time delimiter (not optional). Whitespace or a letter T
    *        (upper or lowercase) which is followed by whitespace. If the letter
    *        T is used, it cannot be preceded by whitespace. No other characters
    *        are allowed.
    *   <li> Time value (optional). If not specified, it is set to 0. It is
    *        specified in HH:MM:SS.sss format, however right side components
    *        can be left off (so HH, HH:MM, HH:MM:SS and HH:MM:SS.sss are all
    *        valid forms). Zero padding is allowed but not required, so hours
    *        can be H or HH, minutes M or MM, seconds S or SS and milliseconds
    *        can be s, ss or sss. Delimiters can only be there if the following
    *        component is included.
    *   <li> Whitespace (optional).
    *   <li> Timezone specification (optional). This is in the form of a sign
    *        (+ or -) then the hours (1 or 2 digits) a colon delimiter and then
    *        minites (1 or 2 digits). Zero padding is allowed but not required.
    *        The minutes can be left off but if so, then the delimiter must
    *        also be left off.  If specified, the instance will be set using
    *        the timezone specified EVEN THOUGH THE TIMEZONE OFFSET IS NOT
    *        STORED (since this is not a datetimetz).
    *   <li> Whitespace (optional).
    * </ol>
    *
    * @param    spec
    *           The date and time specification.
    */
   public datetimetz(String spec)
   throws ErrorConditionException
   {
      instantiateFromStringWorker(spec);
   }

   /**
    * Creates a new instance from the given date and time specification. This
    * is one form of the DATETIME-TZ built-in function. If the input is
    * <code>unknown</code> the instance will be <code>unknown</code>.
    * <p>
    * The specification must match the following format rules:
    * <p>
    * <ol>
    *   <li> Date value (this is not optional).  The string cannot start with
    *        whitespace. The first input is a date value that has at least a
    *        month and day specified or a month, day and year. The values
    *        are parsed in order as specified in SESSION:DATE-FORMAT. Zero
    *        padding is allowed but not required.
    *   <li> Date and time delimiter (not optional). Whitespace or a letter T
    *        (upper or lowercase) which is followed by whitespace. If the letter
    *        T is used, it cannot be preceded by whitespace. No other characters
    *        are allowed.
    *   <li> Time value (optional). If not specified, it is set to 0. It is
    *        specified in HH:MM:SS.sss format, however right side components
    *        can be left off (so HH, HH:MM, HH:MM:SS and HH:MM:SS.sss are all
    *        valid forms). Zero padding is allowed but not required, so hours
    *        can be H or HH, minutes M or MM, seconds S or SS and milliseconds
    *        can be s, ss or sss. Delimiters can only be there if the following
    *        component is included.
    *   <li> Whitespace (optional).
    *   <li> Timezone specification (optional). This is in the form of a sign
    *        (+ or -) then the hours (1 or 2 digits) a colon delimiter and then
    *        minites (1 or 2 digits). Zero padding is allowed but not required.
    *        The minutes can be left off but if so, then the delimiter must
    *        also be left off.  If specified, the instance will be set using
    *        the timezone specified EVEN THOUGH THE TIMEZONE OFFSET IS NOT
    *        STORED (since this is not a datetimetz).
    *   <li> Whitespace (optional).
    * </ol>
    *
    * @param    spec
    *           The date and time specification.
    */
   public datetimetz(character spec)
   throws ErrorConditionException
   {
      if (spec.isUnknown())
      {
         setUnknown();
      }
      else
      {
         instantiateFromStringWorker(spec.toStringMessage());
      }
   }

   /**
    * Creates a new <code>datetimetz</code> instance that represents the 
    * <code>unknown value</code>.
    * 
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public static datetimetz instantiateUnknownDatetimeTz()
   {
      datetimetz ret = new datetimetz();
      ret.setUnknown();
      return ret;
   }

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()
   {
      return instantiateUnknownDatetimeTz();
   }

   /**
    * Creates a new instance of the same type that represents the 
    * default initialized value.
    *
    * @return   An instance that represents the default value.
    */
   public BaseDataType instantiateDefault()
   {
      return instantiateUnknownDatetimeTz();
   }
   
   /**
    * Return the default display format string for this type.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()   
   {
      return defaultFormat;
   }   
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * If the value is not of type <code>datetimetz</code>, the following automatic type
    * conversion will occur:
    * <p>
    * <pre>
    * TODO: implement and document this
    * </pre>
    * <p>
    * This conversion is meant to handle the cases of built-in functions and methods
    * in the 4GL which have polymorphic return types (e.g. DYNAMIC-FUNCTION()).
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BaseDataType value, boolean force)
   {
      if (value instanceof datetimetz)
      {
         assign((datetimetz) value, force);
      }
      else
      {
         // TODO: implement type conversion here for _POLY cases 
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(datetimetz value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         offset = value.offset;
      }
      super.assign((datetime) value, force);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance. The configuration (time-zone) of this
    * instance is not set.  Use {@link #deepAssign} if both state and
    * configuration must be set.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(datetimetz value)
   {
      assign(value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((datetimetz) value, false);
   }
   
   /**
    * Sets the state (data and unknown value) and configuration (time-zone)
    * of this instance based on the state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void deepAssign(datetimetz value)
   {
      super.deepAssign(value);
      offset = value.offset;
   }
   
   /**
    * Obtains the timezone offset in minutes from UTC for this instance. This
    * is the equivalent of the TIMEZONE(datetime-tz) built-in function.
    *
    * @return   The timezone offset for this instance.
    */
   public integer getTimeZoneOffset()
   {
      return isUnknown() ? new integer() : new integer(offset);
   }
   
   /**
    * Returns the legacy type name of the this class.
    * 
    * @return   Always "DATETIME-TZ".
    */
   @Override
   public String getTypeName()
   {
      return "DATETIME-TZ";
   }
   
   /**
    * Get the offset in minutes from UTC for this instance.
    *
    * @return   The offset. If valid it must be between -840 and 840 inclusive.
    *           It should not be invalid.
    */
   public int getOffset()
   {
      return offset;
   }
   
   /**
    * Set the offset in minutes from UTC for this instance.
    *
    * @param    offset
    *           The offset, must be between -840 and 840 inclusive.
    */
   public void setOffset(int offset)
   {
      if (offset < MIN_VALID_TZ || offset > MAX_VALID_TZ)
      {
         String err = "Invalid minute offset in datetime-tz value";
         ErrorManager.recordOrThrowError(11254, err);
         return;
      }
      
      this.offset = offset;
      
      // save this in the super-superclass so that all date operations can
      // honor this
      setTimeZone(getZoneFromOffset(offset));
   }
   
   /**
    * Sets the internal time and offset. This method will validate the
    * parameters and raise an error if they are incorrect. Any input that is
    * <code>unknown</code> will cause this instance to be <code>unknown</code>.
    * 
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    mins
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    secs
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    * @param    offset
    *           The timezone offset in minutes from UTC.  Must be between
    *           -840 and 840 inclusive.
    */
   public void setTime(NumberType hours,
                       NumberType mins,
                       NumberType secs,
                       NumberType millis,
                       NumberType offset)
   {
      if (hours.isUnknown()  ||
          mins.isUnknown()   ||
          secs.isUnknown()   ||
          millis.isUnknown() ||
          offset.isUnknown())
      {
         setUnknown();
      }
      else
      {
         setTime(hours.intValue(),
                 mins.intValue(),
                 secs.intValue(),
                 millis.intValue(),
                 offset.intValue());
      }
   }

   /**
    * Sets the internal time and offset. This method will validate the
    * parameters and raise an error if they are incorrect.
    * 
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    mins
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    secs
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    * @param    offset
    *           The timezone offset in minutes from UTC.  Must be between
    *           -840 and 840 inclusive.
    */
   public void setTime(int hours, int mins, int secs, int millis, int offset)
   {
      // let the superclass set the time value, do this before anything else to
      // allow the error processing to occur first
      setTime(hours, mins, secs, millis);
      
      setOffset(offset);
   }
   
   /**
    * Obtains the text form of this date instance in ISO 8601 standard format.
    * 
    * @return   A string that is the ISO representation of this date 
    *           ("YYYY-MM-DDTHH:MM:SS.SSS+HH:MM") or "?" if
    *           <code>unknown</code>.
    */
   @Override
   protected String getIsoDate()
   {
      if (isUnknown())
      {
         return "?";
      }

      StringBuilder sb = new StringBuilder(super.getIsoDate());
      sb.append('+')
        .append(date.formatNDigits(getTzHours(), 2))
        .append(':')
        .append(date.formatNDigits(getTzMinutes(), 2));

      return sb.toString();
   }

   /**
    * Obtains the minutes of timezone offset.
    * 
    * @return  The minutes of timezone offset.
    */
   public int getTzMinutes()
   {
      return offset % 60;
   }

   /**
    * Obtains the hours of timezone offset.
    * 
    * @return  The hours of timezone offset or unknown value if this 
    *          datetime-tz is unknown.
    */
   public int getTzHours()
   {
      return offset / 60;
   }

   /**
    * Obtains the minutes of timezone offset.
    * 
    * @return  The minutes of timezone offset or unknown value if this 
    *          datetime-tz is unknown.
    */
   public integer getTzMinutesNum()
   {
      return isUnknown() ? new integer() : new integer(getTzMinutes());
   }

   /**
    * Obtains the hours of timezone offset.
    * 
    * @return  The hours of timezone offset or unknown value if this 
    *          datetime-tz is unknown.
    */
   public integer getTzHoursNum()
   {
      return isUnknown() ? new integer() : new integer(getTzHours());
   }


   /**
    * Validate specified values and return error information if they don't consist of correct 
    * datetime  components.
    * 
    * @param   month
    * @param   day
    * @param   year
    * @param   forceFourDigits
    * @param   millis 
    * @param   seconds 
    * @param   minutes 
    * @param   hours 
    * 
    * @return   <code>null</code> if components are valid and error information otherwise.
    */
   public static ErrorEntry validate(int month,
                                     int day,
                                     int year,
                                     boolean forceFourDigits,
                                     int hours,
                                     int minutes,
                                     int seconds,
                                     int millis,
                                     int tz_hours, 
                                     int tz_minutes)
   {
      // check the date part using corresponding static function from date:
      ErrorEntry err = datetime.validate(month, day, year, forceFourDigits, 
                                         hours, minutes, seconds, millis);
      if (err != null)
      {
         return err;
      }

      // now all that remained is to check the time parts:
      if (tz_hours < -12 || tz_hours > 14)
      {
         return new ErrorEntry(80, "The timezone hours of a datetime-tz must be from -12 to 14");
      }
      if (tz_minutes != 0 && tz_minutes != 30 && tz_minutes != 45)
      {
         return new ErrorEntry(80, "The timezone minutes of a datetime-tz must be 0, 30 or 45");
      }
      
      return null;
   }

   /**
    * Parses the given date and time specification and assigns the values into
    * the instance.  The result will be a date and time value that is "local"
    * to the specified timezone (if any) or to the default timezone (if not
    * otherwise specified).
    * <p>
    * The specification must match the following format rules:
    * <p>
    * <ol>
    *   <li> Date value (this is not optional).  The string cannot start with
    *        whitespace. The first input is a date value that has at least a
    *        month and day specified or a month, day and year. The values
    *        are parsed in order as specified in SESSION:DATE-FORMAT. Zero
    *        padding is allowed but not required.
    *   <li> Date and time delimiter (not optional). Whitespace or a letter T
    *        (upper or lowercase) which is followed by whitespace. If the letter
    *        T is used, it cannot be preceded by whitespace. No other characters
    *        are allowed.
    *   <li> Time value (optional). If not specified, it is set to 0. It is
    *        specified in HH:MM:SS.sss format, however right side components
    *        can be left off (so HH, HH:MM, HH:MM:SS and HH:MM:SS.sss are all
    *        valid forms). Zero padding is allowed but not required, so hours
    *        can be H or HH, minutes M or MM, seconds S or SS and milliseconds
    *        can be s, ss or sss. Delimiters can only be there if the following
    *        component is included.  If specified as s or ss, the value is
    *        padded on the RIGHT with zeros to get a 3 digit value.
    *   <li> PROGRESS 4GL BUG ALERT: if the milliseconds value is specified as
    *        a single digit (and only this case) and there is no timezone
    *        specification (but whitespace can follow), then the s value is
    *        used BOTH as a milliseconds value AND as a positive number of
    *        hours in a timezone specification.  So a milliseconds value of
    *        5 would be the equivalent of a "+05:00" specification, while a
    *        value of 05, 50, 005 or 500 would be treated only as a milliseconds
    *        value.  This only occurs for the datetime-tz type
    *   <li> Whitespace (optional).
    *   <li> Timezone specification (optional). This is in the form of a sign
    *        (+ or -) then the hours (1 or 2 digits) a colon delimiter and then
    *        minites (1 or 2 digits). Zero padding is allowed but not required.
    *        The minutes can be left off but if so, then the delimiter must
    *        also be left off.  If specified, the instance will be set using
    *        the timezone specified.
    *   <li> Whitespace (optional).
    * </ol>
    *
    * @param    spec
    *           The date and time specification.
    */
   @Override
   protected void instantiateFromStringWorker(String spec)
   {
      // TODO: implement + make sure all the errors are thrown properly!
      
      // split off the timezone spec and then:
      // offset = date.specToOffset(tzspec);
      
      // I think we need to set the date.zone member and use a worker that
      // uses that value if set OR the nullGC when zone is not set; this would
      // need to be honored in all processing at the date and datetime levels
      
      // use the datetime.instantiateFromStringWorker() to set the date and
      // time from the smaller spec, honoring any zone assignment that was
      // made
   }
}
